from dataclasses import replace
from typing import Optional, Protocol, Tuple

import jax
import jax.numpy as jnp
from pantea.atoms import Structure
from pantea.atoms._structure import _calculate_center_of_mass
from pantea.logger import logger
from pantea.simulation.thermostat import BrendsenThermostat
from pantea.types import Array, Element
from pantea.units import units

KB: float = units.BOLTZMANN_CONSTANT


class PotentialInterface(Protocol):
    def __call__(self, structure: Structure) -> Array:
        ...

    def compute_forces(self, structure: Structure) -> Array:
        ...


@jax.jit
def _get_kinetic_energy(velocities: Array, masses: Array) -> Array:
    return 0.5 * jnp.sum(masses * velocities * velocities)


@jax.jit
def _get_temperature(velocities: Array, masses: Array) -> Array:
    kinetic_energy = _get_kinetic_energy(velocities, masses)
    natoms = velocities.shape[0]
    return 2 * kinetic_energy / (3 * natoms * KB)


@jax.jit
def _get_virial(
    velocities: Array,
    masses: Array,
    positions: Array,
    forces: Array,
) -> Array:
    return (
        2 * _get_kinetic_energy(velocities, masses)
        + (positions * forces).sum()
    )


@jax.jit
def _get_verlet_new_positions(
    positions: Array,
    velocities: Array,
    forces: Array,
    time_step: Array,
) -> Array:
    return (
        positions
        + velocities * time_step
        + 0.5 * forces * time_step * time_step
    )


@jax.jit
def _get_verlet_new_velocities(
    velocities: Array,
    forces: Array,
    new_forces: Array,
    time_step: Array,
) -> Array:
    return velocities + 0.5 * (forces + new_forces) * time_step


def _get_potential_energy(
    potential: PotentialInterface,
    structure: Structure,
) -> Array:
    return potential(structure)


def _compute_forces(
    potential: PotentialInterface,
    structure: Structure,
) -> Array:
    return potential.compute_forces(structure)


class MDSimulator:
    def __init__(
        self,
        potential: PotentialInterface,
        initial_structure: Structure,
        time_step: float,
        initial_velocities: Optional[Array] = None,
        temperature: Optional[float] = None,
        thermostat: Optional[BrendsenThermostat] = None,
        atomic_masses: Optional[Array] = None,
        random_seed: int = 12345,
    ) -> None:
        """
        A molecular dynamics simulator to predict trajectory of particles over time.

        :param potential: input potential function
        :type potential: PotentialInterface
        :param initial_structure: initial structure of atoms
        :type initial_structure: Structure
        :param time_step: time step in Hartree time unit
        :type time_step: float
        :param initial_velocities: Initial atom velocities,
            otherwise it can be randomly generated from the input temperature, defaults to None
        :type initial_velocities: Optional[Array], optional
        :param temperature: input temperature which can be used for generating random velocities
            or setting target temperature for internally initialized thermostat if needed, defaults to None
        :type temperature: Optional[float], optional
        :param thermostat: input thermostat that controls the temperature of the system
            to the desired value, defaults to None
        :type thermostat: Optional[BrendsenThermostat], optional
        :param atomic_masses: atomic mass of atoms in the input structure, defaults to None
        :type atomic_masses: Optional[Array], optional
        :param random_seed: seed for generating random velocities, defaults to 12345
        :type random_seed: int, optional

        This simulator is intended to be used for small systems of atoms during
        the potential training in order to generate new dataset.
        """

        logger.debug(f"Initializing {self.__class__.__name__}")
        self.potential = potential
        self.time_step = time_step
        self.thermostat = thermostat

        self.masses: Array
        if atomic_masses is not None:
            self.masses = atomic_masses.reshape(-1, 1)
        else:
            logger.info("Extracting atomc masses from input structure")
            self.masses = initial_structure.get_masses().reshape(-1, 1)

        self.velocities: Array
        if initial_velocities is not None:
            self.velocities = initial_velocities
        else:
            assert (
                temperature is not None
            ), "At least one of input initial temperature or initial velocities must be given"
            logger.info(f"Generating random velocities ({temperature:0.2f} K)")
            self.velocities = self.generate_random_velocities(
                temperature, masses=self.masses, seed=random_seed
            )

        if (self.thermostat is None) and (temperature is not None):
            logger.info(f"Initializing thermostat ({temperature:0.2f} K)")
            logger.info(
                "Setting default thermostat constant to 100x time step"
            )
            self.thermostat = BrendsenThermostat(
                target_temperature=temperature,
                time_constant=100 * self.time_step,
            )

        self.step: int = 0
        self.elapsed_time: float = 0.0
        self.forces = _compute_forces(self.potential, initial_structure)
        self.temperature = _get_temperature(self.velocities, self.masses)
        self._structure = replace(initial_structure)

    def update(self) -> None:
        """Update parameters for next time step."""
        self.verlet_integration()
        self.step += 1
        self.elapsed_time += self.time_step
        self.temperature = _get_temperature(self.velocities, self.masses)
        if self.thermostat is not None:
            self.velocities = self.thermostat.get_rescaled_velocities(self)
            self.temperature = _get_temperature(self.velocities, self.masses)

    def verlet_integration(self) -> None:
        """Update atom positions and velocities based on Verlet algorithm."""
        arraylike_time_step = jnp.array(self.time_step)
        new_positions = _get_verlet_new_positions(
            self.positions, self.velocities, self.forces, arraylike_time_step
        )
        new_forces = _compute_forces(self.potential, self._structure)
        self.velocities = _get_verlet_new_velocities(
            self.velocities, self.forces, new_forces, arraylike_time_step
        )
        self.forces = new_forces
        self._structure = replace(self._structure, positions=new_positions)

    @classmethod
    def generate_random_velocities(
        cls,
        temperature: float,
        masses: Array,
        seed: int,
    ) -> Array:
        """Generate Maxwell-Boltzmann distributed random velocities."""
        key = jax.random.PRNGKey(seed)
        masses = masses.reshape(-1, 1)
        natoms = masses.shape[0]
        velocities = jax.random.normal(key, shape=(natoms, 3))
        velocities *= jnp.sqrt(
            temperature / _get_temperature(velocities, masses)
        )
        velocities -= _calculate_center_of_mass(velocities, masses)
        return velocities

    def repr_physical_params(self) -> str:
        """Represent current physical parameters."""
        return (
            f"{self.step:<10} "
            f"time[ps]:{units.TO_PICO_SECOND * self.elapsed_time:<10.5f} "
            f"Temp[K]:{self.temperature:<10.5f} "
            f"Etot[Ha]:{self.get_total_energy():<15.10f} "
            f"Epot[Ha]:{self.get_potential_energy():<15.10f} "
            f"Pres[kb]:{self.get_pressure() * units.TO_KILO_BAR:<10.5f}"
            if self._structure.box
            else ""
        )

    @property
    def positions(self) -> Array:
        return self._structure.positions

    @property
    def natoms(self) -> int:
        return self._structure.natoms

    def get_elements(self) -> Tuple[Element]:
        return self._structure.get_elements()

    def get_structure(self) -> Structure:
        return replace(
            self._structure,
            forces=self.forces,
            total_energy=self.potential(self._structure),
        )

    def get_pressure(self) -> Array:
        assert (
            self._structure.box
        ), "Calulating pressure... input structure must have PBC box"
        volume = self._structure.box.volume
        virial = _get_virial(
            self.velocities,
            self.masses,
            self.positions,
            self.forces,
        )
        return virial / (3.0 * volume)  # type: ignore

    def get_center_of_mass_velocity(self) -> Array:
        return _calculate_center_of_mass(self.velocities, self.masses)

    def get_center_of_mass_position(self) -> Array:
        return _calculate_center_of_mass(self.positions, self.masses)

    def get_potential_energy(self) -> Array:
        return _get_potential_energy(self.potential, self._structure)

    def get_kinetic_energy(self) -> Array:
        return _get_kinetic_energy(self.velocities, self.masses)

    def get_total_energy(self) -> Array:
        return self.get_potential_energy() + self.get_kinetic_energy()



import itertools
from dataclasses import dataclass
from typing import Optional, Tuple

import jax.numpy as jnp
from pantea.atoms.structure import Structure
from pantea.descriptors.acsf._acsf import (
    _calculate_descriptor,
    _calculate_grad_descriptor_per_atom,
)
from pantea.descriptors.acsf.angular import AngularSymmetryFunction
from pantea.descriptors.acsf.radial import RadialSymmetryFunction
from pantea.descriptors.acsf.symmetry import (
    BaseSymmetryFunction,
    EnvironmentElements,
)
from pantea.descriptors.descriptor import DescriptorInterface
from pantea.logger import logger
from pantea.pytree import BaseJaxPytreeDataClass, register_jax_pytree_node
from pantea.types import Array, Element


@dataclass
class ACSF(BaseJaxPytreeDataClass, DescriptorInterface):
    """
    Atom-centered Symmetry Function (`ACSF`_) descriptor captures
    information about the distribution of neighboring
    atoms around a central atom by considering both radial (two-body) and angular
    (three-body) symmetry functions. Radial symmetry functions describe the distances
    between the central atom and its neighbors, while angular symmetry functions capture
    the angles formed by the central atom with pairs of neighboring atoms.

    The ACSF descriptor is computed by summing the contributions of the `symmetry functions`
    for all neighboring atoms within a specified cutoff distance. The values obtained from
    these calculations represent a fingerprint of the local atomic environment and can
    be used in various machine learning potentials.


    Example
    -------

    .. code-block:: python
        :linenos:

        from pantea.descriptors.acsf import ACSF, G2, G3, G9, CutoffFunction

        # Create an instance of ACSF for `O` element
        acsf = ACSF('O')

        # Set cutoff function and symmetry functions
        cfn = CutoffFunction(12.0)
        g2 = G2(cfn, eta=0.5, r_shift=0.0)
        g3 = G3(cfn, eta=0.001, zeta=2.0, lambda0=1.0, r_shift=12.0)

        # Add different symmetry functions
        acsf.add(g2, 'O')
        acsf.add(g2, 'H')
        acsf.add(g3, 'H', 'H')
        acsf.add(g3, 'H', 'O')

        print(acsf)

    Output:

    .. code-block:: bash

        ACSF(central_element='O', num_symmetry_functions=4, r_cutoff=12.0)


    .. _ACSF: https://compphysvienna.github.io/n2p2/topics/descriptors.html?highlight=symmetry%20function#

    """

    central_element: str
    radial_symmetry_functions: Tuple[Tuple[EnvironmentElements, RadialSymmetryFunction]] = tuple()  # type: ignore
    angular_symmetry_functions: Tuple[Tuple[EnvironmentElements, AngularSymmetryFunction]] = tuple()  # type: ignore

    def add(
        self,
        symmetry_function: BaseSymmetryFunction,
        neighbor_element_j: Element,
        neighbor_element_k: Optional[Element] = None,
    ) -> None:
        """Add the input symmetry function to the list of ACSFs."""
        if isinstance(symmetry_function, RadialSymmetryFunction):
            self.radial_symmetry_functions = self.radial_symmetry_functions + (
                (
                    EnvironmentElements(
                        self.central_element, neighbor_element_j
                    ),
                    symmetry_function,
                ),
            )  # type: ignore
        elif isinstance(symmetry_function, AngularSymmetryFunction):
            self.angular_symmetry_functions = self.angular_symmetry_functions + (
                (
                    EnvironmentElements(
                        self.central_element, neighbor_element_j, neighbor_element_k  # type: ignore
                    ),
                    symmetry_function,
                ),
            )
        else:
            logger.error(
                f"Unknown symmetry function type {symmetry_function}",
                exception=TypeError,
            )

    def __call__(
        self,
        structure: Structure,
        atom_indices: Optional[Array] = None,
    ) -> Array:
        """
        Calculate descriptor values for the input given structure and atom index.

        :param structure: input structure instance
        :type structure: Structure
        :param atom_indices: atom indices, defaults select all atom indices
        of type the central element of the descriptor.
        :type atom_indices: Optional[Array], optional
        :return: descriptor values
        :rtype: Array
        """

        if self.num_symmetry_functions == 0:
            logger.warning("No symmetry function was found")

        if atom_indices is None:
            atom_indices = structure.select(self.central_element)
        else:
            atom_indices = jnp.atleast_1d(atom_indices)  # type: ignore
            # Check aid atom type match the central element
            if not jnp.all(
                structure.element_map.element_to_atom_type[
                    self.central_element
                ]
                == structure.atom_types[atom_indices]
            ):
                logger.error(
                    f"Inconsistent central element '{self.central_element}': "
                    f" input atom indices={atom_indices}"
                    f" (atom_types='{int(structure.atom_types[atom_indices])}')",
                    exception=ValueError,
                )

        return _calculate_descriptor(
            self,
            structure.positions[atom_indices],
            structure.positions,
            structure.atom_types,
            structure.lattice,
            structure.element_map.element_to_atom_type,
        )

    def grad(
        self,
        structure: Structure,
        atom_index: int,
    ) -> Array:
        """
        Compute gradient of ACSF descriptor respect to atom position (x, y, and z).

        :param structure: input Structure instance
        :param atom_index: atom index in Structure [0, natoms)
        :return: gradient of the descriptor value respect to the atom position
        """
        if not (0 <= atom_index < structure.natoms):
            logger.error(
                f"Unexpected {atom_index=}."
                f" The index must be between [0, {structure.natoms})",
                ValueError,
            )

        return _calculate_grad_descriptor_per_atom(
            self,
            structure.positions[
                atom_index
            ],  # must be a single atom position shape=(1, 3)
            structure.positions,
            structure.atom_types,
            structure.lattice,
            structure.element_map.element_to_atom_type,
        )

    @property
    def num_radial_symmetry_functions(self) -> int:
        """Return number of `two-body` (radial) symmetry functions."""
        return len(self.radial_symmetry_functions)

    @property
    def num_angular_symmetry_functions(self) -> int:
        """Return number of `three-body` (angular) symmetry functions."""
        return len(self.angular_symmetry_functions)

    @property
    def num_symmetry_functions(self) -> int:
        """Return the total (`two-body` and `tree-body`) number of symmetry functions."""
        return (
            self.num_radial_symmetry_functions
            + self.num_angular_symmetry_functions
        )

    @property
    def num_descriptors(self) -> int:
        return self.num_symmetry_functions

    @property
    def r_cutoff(self) -> float:  # type: ignore
        """Return the maximum cutoff radius for list of the symmetry functions."""
        return max(
            [
                cfn.r_cutoff
                for (_, cfn) in itertools.chain(
                    *[
                        self.radial_symmetry_functions,
                        self.angular_symmetry_functions,
                    ]
                )
            ]
        )

    def __hash__(self) -> int:
        """Enforce to use the parent class's hash method (JIT)."""
        return super().__hash__()

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(central_element='{self.central_element}'"
            f", symmetry_functions={self.num_symmetry_functions})"
        )


register_jax_pytree_node(ACSF)



#!/usr/bin/env python

"""The setup script."""

from setuptools import find_packages, setup

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = [
    "Click>=7.0",
    "tqdm>=4.6",
    "ase>=3.2",
    "frozendict>=2.3",
    "flax>=0.6",
    "pydantic>=1.10",
    "pandas>=1.5.3",
]

test_requirements = [
    "pytest>=3",
]

_dict = {}
with open("pantea/_version.py") as f:
    exec(f.read(), _dict)
__version__ = _dict["__version__"]

setup(
    author="Hossein Ghorbanfekr",
    author_email="hgh.comphys@gmail.com",
    python_requires=">=3.8",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.8",
    ],
    description="A Python package for developing machine learning interatomic potentials, based on JAX.",
    entry_points={
        "console_scripts": [
            "pantea=pantea.cli:main",
        ],
    },
    install_requires=requirements,
    license="GNU General Public License v3",
    long_description=readme,
    include_package_data=True,
    keywords="pantea",
    name="pantea",
    packages=find_packages(include=["pantea", "pantea.*"]),
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/hghcomphys/pantea",
    version=__version__,
    zip_safe=False,
)



from typing import Protocol

import jax.numpy as jnp
from pantea.types import Array


class MDSimulatorInterface(Protocol):
    temperature: float
    time_step: float
    velocities: Array


class BrendsenThermostat:
    """Control simulation temperature using Brendsen algorithm."""

    def __init__(
        self,
        target_temperature: float,
        time_constant: float,
    ) -> None:
        self.target_temperature = target_temperature
        self.time_constant = time_constant

    def get_rescaled_velocities(
        self,
        simulator: MDSimulatorInterface,
    ) -> Array:
        scaling_factor = jnp.sqrt(
            1.0
            + (simulator.time_step / self.time_constant)
            * (simulator.temperature / self.target_temperature - 1.0)
        )
        return simulator.velocities / scaling_factor



from functools import partial
from typing import Dict, Protocol

import jax.numpy as jnp
from frozendict import frozendict
from jax import jit
from pantea.atoms.structure import Inputs
from pantea.descriptors import Scaler
from pantea.descriptors.acsf._acsf import _calculate_descriptor
from pantea.descriptors.descriptor import DescriptorInterface
from pantea.models import NeuralNetworkModel
from pantea.types import Array, Element


class AtomicPotentialInterface(Protocol):
    """An interface for AtomicPotential."""

    descriptor: DescriptorInterface
    scaler: Scaler
    model: NeuralNetworkModel


@partial(jit, static_argnums=(0,))
def _compute_atomic_energy(
    atomic_potential: AtomicPotentialInterface,
    positions: Array,
    params: frozendict,
    inputs: Inputs,
) -> Array:
    """Compute model output per-atom energy."""
    x = _calculate_descriptor(
        atomic_potential.descriptor,
        positions,
        inputs.positions,
        inputs.atom_types,
        inputs.lattice,
        inputs.emap,
    )
    x = atomic_potential.scaler(x)
    atomic_energy = atomic_potential.model.apply({"params": params}, x)
    return atomic_energy  # type: ignore


@partial(jit, static_argnums=(0,))
def _energy_fn(
    atomic_potential: Dict[Element, AtomicPotentialInterface],
    positions: Dict[Element, Array],
    params: Dict[Element, frozendict],
    inputs: Dict[Element, Inputs],
) -> Array:
    """
    A helper function that allows to calculate gradient of the NNP total energy
    respect to the atom positions (for each element).
    """
    elements: list[Element] = list(inputs.keys())
    total_energy: Array = jnp.array(0.0)
    for element in elements:
        atomic_energy = _compute_atomic_energy(
            atomic_potential[element],
            positions[element],
            params[element],
            inputs[element],
        )
        total_energy += jnp.sum(atomic_energy)
    return total_energy


_compute_energy = _energy_fn



from __future__ import annotations

from typing import Callable, Optional, Tuple

import jax
import jax.numpy as jnp
from pantea.atoms.box import _apply_pbc
from pantea.types import Array


# @jax.jit
def _calculate_distances_per_atom(
    atom_positions: Array,
    neighbor_positions: Array,
    lattice: Optional[Array] = None,
) -> Tuple[Array, Array]:
    """Calculate distances between a single atom and neighboring atoms."""
    dx: Array = atom_positions - neighbor_positions
    if lattice is not None:
        dx = _apply_pbc(dx, lattice)
    # Fix NaN in gradient of np.linalg.norm
    # see https://github.com/google/jax/issues/3058
    is_zero = dx.sum(axis=1, keepdims=True) == 0.0
    dx_ = jnp.where(is_zero, jnp.ones_like(dx), dx)
    dist = jnp.linalg.norm(dx_, ord=2, axis=1)
    dist = jnp.where(jnp.squeeze(is_zero), 0.0, dist)
    return dist, dx


_vmap_calculate_distances: Callable = jax.vmap(
    _calculate_distances_per_atom,
    in_axes=(0, None, None),
)


@jax.jit
def _calculate_distances(
    atom_positions: Array,
    neighbor_positions: Array,
    lattice: Optional[Array] = None,
) -> Tuple[Array, Array]:
    """Calculate an array of distances between multiple atoms
    and the neighbors (using `jax.vmap`)."""
    return _vmap_calculate_distances(
        atom_positions, neighbor_positions, lattice
    )


@jax.jit
def _calculate_center_of_mass(array: Array, masses: Array) -> Array:
    return jnp.sum(masses * array, axis=0) / jnp.sum(masses)



import pickle
from dataclasses import field
from pathlib import Path
from typing import Callable, List, Tuple

from flax import linen as nn
from frozendict import frozendict
from pantea.logger import logger
from pantea.models.model import ModelInterface
from pantea.models.nn.activation import _activation_function_map
from pantea.types import Array, Dtype, default_dtype


class NeuralNetworkModel(nn.Module, ModelInterface):
    """Neural network model that outputs energy."""

    hidden_layers: Tuple[Tuple[int, str], ...]
    output_layer: Tuple[int, str] = (1, "identity")
    param_dtype: Dtype = field(default_factory=lambda: default_dtype.FLOATX)
    kernel_initializer: Callable = nn.initializers.lecun_normal()
    # bias_initializer: Callable = nn.initializers.zeros

    def setup(self) -> None:
        """Initialize neural network model."""
        self.layers: List = self.create_network()

    def create_layer(self, features: int) -> nn.Dense:
        """
        Create a dense layer and initialize the weights and biases
        (see `here <https://aiqm.github.io/torchani/examples/nnp_training.html#training-example>`_).
        """
        return nn.Dense(
            features,
            param_dtype=self.param_dtype,
            kernel_init=self.kernel_initializer,
            bias_init=nn.initializers.zeros,
        )

    def create_network(self) -> List:
        """Create a neural network as stack of dense layers and activation functions."""

        layers: List = list()
        # Hidden layers
        for out_size, af_type in self.hidden_layers:
            layers.append(self.create_layer(out_size))
            layers.append(_activation_function_map[af_type])
        # Output layer
        layers.append(self.create_layer(self.output_layer[0]))
        layers.append(_activation_function_map[self.output_layer[1]])
        return layers

    def __call__(self, inputs: Array) -> Array:
        """Compute energy."""
        x = inputs
        for layer in self.layers:
            x = layer(x)
        return x

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(hidden_layers={self.hidden_layers}"
            # f", output_layer={self.output_layer}"
            f", param_dtype={self.param_dtype.dtype}"  # type: ignore
            ")"
        )

    def save(self, filename: Path, params: frozendict) -> None:
        """Save model weights."""
        file = str(Path(filename))
        logger.debug(f"Saving model weights into '{file}'")
        with open(file, "wb") as handle:
            pickle.dump(params, handle)

    def load(self, filename: Path) -> frozendict:
        """Load model weights."""
        file = str(Path(filename))
        logger.debug(f"Loading model weights from '{file}'")
        with open(file, "rb") as handle:
            params: frozendict = pickle.load(handle)
        return params



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import Callable, Mapping, Optional, Type

import jax.numpy as jnp
from pantea.logger import logger
from pantea.types import Array


class ErrorMetric(metaclass=ABCMeta):
    """A base error metric class."""

    def __init__(self) -> None:
        """Initialize base error metric."""

        def mse(*, prediction: Array, target: Array) -> Array:
            return ((target - prediction) ** 2).mean()

        self._mse_metric: Callable[..., Array] = mse

    @classmethod
    def create(cls, metric_type: str) -> ErrorMetric:
        """
        Create the given type of error metric.

        :param metric_type: MSE, RMSE, MSEpa, EMSEpa
        :return: An instance of desired error metric
        """
        _map_error_metric: Mapping[str, Type[ErrorMetric]] = {
            "MSE": MSE,
            "RMSE": RMSE,
            "MSEpa": MSEpa,
            "RMSEpa": RMSEpa,
            # Any new defined metrics must be added here.
        }
        try:
            error_metric = _map_error_metric[metric_type]()
        except KeyError:
            logger.error(
                f"Unknown error metric '{metric_type}'",
                exception=KeyError,
            )
        return error_metric  # type: ignore

    @abstractmethod
    def __call__(self, prediction: Array, target: Array, factor=None) -> Array:
        ...

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"


class MSE(ErrorMetric):
    """Mean squared error."""

    def __call__(
        self, prediction: Array, target: Array, factor: Optional[float] = None
    ) -> Array:
        return self._mse_metric(prediction=prediction, target=target)


class RMSE(MSE):
    """Root mean squared error."""

    def __call__(
        self, prediction: Array, target: Array, factor: Optional[float] = None
    ) -> Array:
        return jnp.sqrt(self._mse_metric(prediction=prediction, target=target))


class MSEpa(MSE):
    """
    Mean squared error per atom.
    MSE of energy **per atom**
    MSE of force
    """

    def __call__(
        self, prediction: Array, target: Array, factor: float = 1.0
    ) -> Array:
        return self._mse_metric(prediction=prediction, target=target) / factor


class RMSEpa(RMSE):
    """
    Root mean squared error per atom.
    RMSE of energy **per atom**
    RMSE of force
    """

    def __call__(
        self, prediction: Array, target: Array, factor: float = 1.0
    ) -> Array:
        return (
            jnp.sqrt(self._mse_metric(prediction=prediction, target=target))
            / factor
        )



from __future__ import annotations

import math
import random
from collections import defaultdict
from typing import Any, Callable, Dict, List, Tuple

import jax.numpy as jnp
import numpy as np
import optax
from flax.training.train_state import TrainState
from frozendict import frozendict
from jax import value_and_grad
from optax import GradientTransformation
from pantea.atoms.structure import Structure
from pantea.datasets.dataset import DatasetInterface
from pantea.logger import logger
from pantea.potentials._energy import _energy_fn
from pantea.potentials._force import _compute_force
from pantea.potentials.nnp.metrics import ErrorMetric
from pantea.potentials.nnp.nnp import (
    NeuralNetworkPotentialInterface as PotentialInterface,
)
from pantea.potentials.nnp.settings import (
    NeuralNetworkPotentialSettings as PotentialSettings,
)
from pantea.types import Array, Element
from tqdm import tqdm


def _mse_loss(*, logits: Array, targets: Array) -> Array:
    return ((targets - logits) ** 2).mean()


# @dataclass
class GradientDescentUpdater:
    """
    A trainer class to fit a generic NNP potential weights using target values of the total
    energy and force components.

    See https://pytorch.org/tutorials/beginner/introyt/trainingyt.html
    """

    # potential: PotentialInterface
    # settings: PotentialSettings
    # criterion: Callable = field(default_factory=lambda: mse_loss)
    # error_metric: ErrorMetric
    # optimizer: Dict[Element, Any]

    def __init__(self, potential: PotentialInterface) -> None:
        """Initialize potential."""
        self.potential: PotentialInterface = potential
        self.criterion: Callable[..., Array] = _mse_loss
        self.error_metric: ErrorMetric = ErrorMetric.create(
            self.potential.settings.main_error_metric
        )
        self._init_parameters()
        self._init_optimizer()

    def _init_parameters(self) -> None:
        """Set required parameters from the potential settings."""
        settings: PotentialSettings = self.potential.settings
        self.gradient_type: str = settings.gradient_type
        self.beta: float = settings.force_weight
        self.force_fraction: float = settings.short_force_fraction

    def _init_optimizer(self) -> None:
        """Create optimizer using the potential settings."""
        settings: PotentialSettings = self.potential.settings
        if self.gradient_type == "adam":
            self.optimizer: GradientTransformation = optax.adamw(
                learning_rate=settings.gradient_adam_eta,
                b1=settings.gradient_adam_beta1,
                b2=settings.gradient_adam_beta2,
                eps=settings.gradient_adam_epsilon,
                weight_decay=settings.gradient_adam_weight_decay,
            )
        else:  # TODO: self.settings.gradient_type == "fixed_step":
            logger.error(
                f"Gradient type {settings.gradient_type} is not implemented yet",
                exception=TypeError,
            )
        # Single global optimizer or multiple optimizers:
        # return {element: optimizer for element in self.elements}

    def _init_train_state(self) -> Dict[Element, TrainState]:
        """Initialize train state for each element."""

        def generate_train_state():
            for element in self.potential.elements:
                yield element, TrainState.create(
                    apply_fn=self.potential.atomic_potential[
                        element
                    ].model.apply,
                    params=self.potential.model_params[element],
                    tx=self.optimizer,  # [element]?
                )

        return {element: state for element, state in generate_train_state()}

    # ------------------------------------------------------------------------

    def fit(self, dataset: DatasetInterface, **kwargs):
        """Train potential."""
        batch_size: int = kwargs.get("batch_size", 1)
        steps: int = kwargs.get("steps", math.ceil(len(dataset) / batch_size))
        epochs: int = kwargs.get("epochs", 10)

        states: Dict[Element, TrainState] = self._init_train_state()
        history = defaultdict(list)

        # Loop over epochs
        for epoch in range(epochs):
            print(f"[Epoch {epoch+1} of {epochs}]")

            loss_per_epoch: Array = jnp.array(0.0)
            loss_energy_per_epoch: Array = jnp.array(0.0)
            loss_force_per_epoch: Array = jnp.array(0.0)
            num_updates_per_epoch: int = 0

            # Loop over batches
            for _ in tqdm(range(steps)):
                structures: List[Structure] = random.choices(
                    dataset, k=batch_size
                )
                xbatch = tuple(
                    structure.get_per_element_inputs()
                    for structure in structures
                )
                ybatch = tuple(
                    (
                        structure.total_energy,
                        structure.get_per_element_forces(),
                    )
                    for structure in structures
                )

                batch = xbatch, ybatch
                states, metrics = self.train_step(states, batch)

                loss_per_epoch += metrics["loss"]
                loss_energy_per_epoch += metrics["loss_energy"]
                loss_force_per_epoch += metrics["loss_force"]
                num_updates_per_epoch += 1

            loss_per_epoch /= num_updates_per_epoch
            loss_energy_per_epoch /= num_updates_per_epoch
            loss_force_per_epoch /= num_updates_per_epoch

            self._update_model_params(states)

            print(
                f"training loss:{float(loss_per_epoch): 0.7f}"
                f", loss_energy:{float(loss_energy_per_epoch): 0.7f}"
                f", loss_force:{float(loss_force_per_epoch): 0.7f}"
            )
            history["epoch"].append(epoch)
            history["loss"].append(loss_per_epoch)

        return history

    # @partial(jit, static_argnums=(0,))
    def train_step(
        self,
        states: Dict[Element, TrainState],
        batch: Tuple,
    ) -> Tuple:
        """Train potential on a batch of data."""

        def loss_fn(params: Dict[Element, frozendict]) -> Tuple[Array, Any]:
            """Loss function."""
            xbatch, ybatch = batch
            batch_size = len(xbatch)

            loss_energy_per_batch: Array = jnp.array(0.0)
            loss_force_per_batch: Array = jnp.array(0.0)

            for inputs, (true_energy, true_forces) in zip(xbatch, ybatch):
                positions = {
                    element: input.atom_positions
                    for element, input in inputs.items()
                }
                natoms: int = sum(
                    array.shape[0] for array in positions.values()
                )

                if np.random.rand() < self.force_fraction:
                    # ------ Force ------
                    forces = _compute_force(
                        frozendict(self.potential.atomic_potential),
                        positions,
                        params,
                        inputs,
                    )
                    elements = true_forces.keys()
                    loss_force = jnp.array(0.0)
                    for element in elements:
                        loss_force += self.criterion(
                            logits=forces[element],
                            targets=true_forces[element],
                        )
                    loss_force /= len(forces)
                    loss_force_per_batch += self.beta * self.beta * loss_force

                else:
                    # ------ energy ------
                    energy = _energy_fn(
                        frozendict(self.potential.atomic_potential),
                        positions,
                        params,
                        inputs,
                    )
                    loss_energy = (
                        self.criterion(logits=energy, targets=true_energy)
                        / natoms
                    )
                    loss_energy_per_batch += loss_energy

            loss_energy_per_batch /= batch_size
            loss_force_per_batch /= batch_size
            loss_per_batch = loss_energy_per_batch + loss_force_per_batch

            return loss_per_batch, (
                loss_energy_per_batch,
                loss_force_per_batch,
            )

        value_and_grad_fn = value_and_grad(loss_fn, has_aux=True)

        (loss, (loss_energy, loss_force)), grads = value_and_grad_fn(
            {element: state.params for element, state in states.items()}
        )
        states = {
            element: states[element].apply_gradients(grads=grads[element])
            for element in states.keys()
        }

        metrics = {
            "loss": loss,
            "loss_energy": loss_energy,
            "loss_force": loss_force,
        }
        return states, metrics

    def _update_model_params(self, states: Dict[Element, TrainState]) -> None:
        """Update model params for all the elements."""
        for element, train_state in states.items():
            self.potential.model_params[element] = train_state.params  # type: ignore

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(\npotential={self.potential}, \noptimizer={self.optimizer}, "
            f"\ncriterion={self.criterion}, \nerror_metric={self.error_metric})"
        )



from typing import Any, Dict, List, Union

import jax.numpy as jnp
from pantea.atoms.structure import Structure
from pantea.types import Array


def compare(
    structure1: Structure,
    structure2: Structure,
    errors: Union[str, List] = "RMSEpa",
    return_difference: bool = False,
) -> Dict[str, Array]:
    """
    Compare the `force` and `total energy` values
    between two input structures and returning the desired errors metrics.

    :param structure1: first structure
    :param structure2: second structure
    :param error: a list of error metrics including `RMSE`, `RMSEpa`, `MSE`, and `MSEpa`. Defaults to [`RMSEpa`]
    :param return_difference: whether return energy and force array differences or not, defaults to False
    :return: a dictionary of error metrics.
    """
    assert all(
        structure1.atom_types == structure2.atom_types
    ), "Expected similar structures with the same atom types."

    result: Dict[str, Any] = dict()
    frc_diff: Array = structure1.forces - structure2.forces
    eng_diff: Array = structure1.total_energy - structure2.total_energy
    errors = [errors] if isinstance(errors, str) else errors
    print(f"Comparing two structures, error metrics: {', '.join(errors)}")
    errors = [x.lower() for x in errors]

    if "rmse" in errors:
        result["force_RMSE"] = jnp.sqrt(jnp.mean(frc_diff**2))
        result["energy_RMSE"] = jnp.sqrt(jnp.mean(eng_diff**2))
    if "rmsepa" in errors:
        result["force_RMSEpa"] = jnp.sqrt(jnp.mean(frc_diff**2))
        result["energy_RMSEpa"] = (
            jnp.sqrt(jnp.mean(eng_diff**2)) / structure1.natoms
        )
    if "mse" in errors:
        result["force_MSE"] = jnp.mean(frc_diff**2)
        result["energy_MSE"] = jnp.mean(eng_diff**2)
    if "msepa" in errors:
        result["force_MSEpa"] = jnp.mean(frc_diff**2)
        result["energy_MSEpa"] = jnp.mean(eng_diff**2) / structure1.natoms
    if return_difference:
        result["frc_diff"] = frc_diff
        result["eng_diff"] = eng_diff

    return result



import os
from functools import partial
from typing import Tuple, cast

import jax
import jax.numpy as jnp
from pantea.atoms.neighbor import Neighbor
from pantea.atoms.structure import Structure
from pantea.types import Array

os.environ["JAX_ENABLE_X64"] = "1"
os.environ["JAX_PLATFORM_NAME"] = "cpu"


# @partial(jax.jit, static_argnums=0)
def _compute_pair_energies(obj, r: Array) -> Array:
    term = obj.sigma / r
    term6 = term**6
    return 4.0 * obj.epsilon * term6 * (term6 - 1.0)


# @partial(jax.jit, static_argnums=0)
def _compute_pair_forces(obj, r: Array, R: Array) -> Array:
    term = obj.sigma / r
    term6 = term**6
    force_factor = -24.0 * obj.epsilon / (r * r) * term6 * (2.0 * term6 - 1.0)
    return jnp.expand_dims(force_factor, axis=-1) * R


def _get_neighbor_data(
    structure: Structure,
    r_cutoff: float,
) -> Tuple[Array, Array, Array]:
    if structure.neighbor is None:
        structure.update_neighbor(r_cutoff)
    neighbor = cast(Neighbor, structure.neighbor)
    return (neighbor.masks, neighbor.rij, neighbor.Rij)


class LJPotential:
    """An implementation of Lennard-Jones potential."""

    def __init__(
        self,
        sigma: float,
        epsilon: float,
        r_cutoff: float,
    ) -> None:
        self.sigma = sigma
        self.epsilon = epsilon
        self.r_cutoff = r_cutoff

    @partial(jax.jit, static_argnums=0)
    def __call__(self, structure: Structure) -> Array:
        """Compute total energy."""
        masks, rij, _ = _get_neighbor_data(structure, self.r_cutoff)
        pair_energies = _compute_pair_energies(self, rij)
        return 0.5 * jnp.where(masks, pair_energies, 0.0).sum()  # type: ignore

    @partial(jax.jit, static_argnums=0)
    def compute_forces(self, structure: Structure) -> Array:
        """Compute force component for each atoms."""
        masks, rij, Rij = _get_neighbor_data(structure, self.r_cutoff)
        pair_forces = jnp.where(
            jnp.expand_dims(masks, axis=-1),
            _compute_pair_forces(self, rij, Rij),
            jnp.zeros_like(Rij),
        )
        return jnp.sum(pair_forces, axis=1)  # type: ignore
